import pytest
import sys

import numpy as np

import pycollimator as collimator
from pycollimator.error import NotFoundError, CollimatorApiError

from tests import get_model_name


@pytest.mark.parametrize("name", ["test_004"])
def test_slice_results_by_path(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"

    results = sim.get_results()
    df = results.to_pandas()
    assert len(df.columns) == 13

    with pytest.raises(NotFoundError):
        results["path that does not exist"]
    with pytest.raises(NotFoundError):
        results.to_pandas("path that does not exist")

    # Read columns from data frame
    adder_full_path = "Group_1.Group_0.SquareRoot_0.out_0"
    assert len(df[adder_full_path]) > 0

    top_level_group_full_path = "Group_1.Outport_0"
    assert len(df[top_level_group_full_path]) > 0

    # Dataframe doesn't recognize paths without ports but results do.
    top_level_group = "Group_1"
    assert len(results[top_level_group]) > 0

    nested_group = "Group_1.Group_0"
    assert len(results[nested_group]) > 0

    # Check find by type constructs path correctly
    sqrt = sim.model.find_block(type="SquareRoot")
    assert sqrt.path == "Group_1.Group_0.SquareRoot_0"
    assert len(results[sqrt]) > 0

    # All columns from the dataframe should be in the results
    for col in results.columns:
        assert len(results[col]) > 0


@pytest.mark.parametrize("name", ["test_004"])
def test_request_results_by_path(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"

    base_results = sim.get_results().to_pandas()
    assert len(base_results.columns) == 13

    # Request specific signal data as single string
    adder_full_path = "Group_1.Group_0.SquareRoot_0.out_0"
    top_level_group_full_path = "Group_1.Outport_0"

    results = sim.get_results(adder_full_path).to_pandas()
    assert results.columns.values.tolist() == [adder_full_path]
    assert np.isclose(results[adder_full_path], base_results[adder_full_path]).all()

    results = sim.get_results(top_level_group_full_path).to_pandas()
    assert results.columns.values.tolist() == [top_level_group_full_path]
    assert np.isclose(results[top_level_group_full_path], base_results[top_level_group_full_path]).all()

    # Request single path
    results = sim.get_results([top_level_group_full_path]).to_pandas()
    assert results.columns.values.tolist() == [top_level_group_full_path]
    assert np.isclose(results[top_level_group_full_path], base_results[top_level_group_full_path]).all()

    # Request data for multiple signals
    multiple_signals = [adder_full_path, top_level_group_full_path]
    results = sim.get_results(multiple_signals).to_pandas()
    assert results.columns.values.tolist() == multiple_signals
    assert np.isclose(results[adder_full_path], base_results[adder_full_path]).all()
    assert np.isclose(results[top_level_group_full_path], base_results[top_level_group_full_path]).all()

    # Raise error for invalid signal path
    with pytest.raises(CollimatorApiError):
        results = sim.get_results(["path that does not exist"])
    # Raise error if there is any invalid signal path
    with pytest.raises(CollimatorApiError):
        results = sim.get_results([*multiple_signals, "path that does not exist"])


if __name__ == "__main__":
    sys.exit(pytest.main([__file__]))



import random
import numpy as np
import pandas as pd
import pytest
import sys

import pycollimator as collimator
from pycollimator.error import NotFoundError

from tests import get_model_name


@pytest.mark.parametrize("name", ["test_001"])
def test_simulation_simple(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)
    assert model.name.startswith(model_name)

    sim = collimator.run_simulation(model, wait=False)
    assert isinstance(sim, collimator.simulations.Simulation)

    sim.wait()
    assert sim.status.lower() == "completed"

    # Get the results
    results = sim.results
    assert isinstance(results, collimator.simulations.SimulationResults)

    df = results.to_pandas()
    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0

    # Get results data from the Test and Control output blocks
    test_output_blk = model.find_block("TestOutput")
    control_output_blk = model.find_block("ControlOutput")

    test_data = results[test_output_blk.name]
    control_data = results[control_output_blk.name]
    assert len(test_data) == len(control_data)
    assert len(test_data) > 0

    # Verify numerical results
    test_data = test_data.to_numpy()
    control_data = control_data.to_numpy()
    assert np.isclose(test_data, control_data).all()


# TODO: clean this up. there's test_configs which has some overlap
@pytest.mark.xfail(reason="Simulation-time configuration is not implemented yet")
@pytest.mark.parametrize("name", ["test_001"])
def test_simulation_configuration(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    stop_time = model.configuration.stop_time * 0.5
    configuration = {"stop_time": stop_time}
    sim = collimator.run_simulation(model, configuration=configuration, wait=True)
    assert sim.status.lower() == "completed"

    results = sim.results
    df = results.to_pandas()
    assert len(df["TestOutput.out_0"]) > 0

    # Check that we ran to the specified stop time
    assert np.isclose(df["TestOutput.out_0"].index[-1], stop_time)


@pytest.mark.parametrize("name", ["test_002"])
def test_simulation_parameters(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    assert len(model.get_parameters().keys()) > 0
    old_value = model.parameters["ParamValue"].value

    # Set param with dict access
    new_value = random.random()
    model.parameters["ParamValue"] = new_value

    with pytest.raises(NotFoundError):
        model.parameters["NewParameter"]
    # Guard against accidental parameter creation
    with pytest.raises(NotFoundError):
        model.parameters["NewParameter"] = 0

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"

    # Check the parameter value did not change: parameters aren't supposed to be changed
    model.reload()
    assert model.parameters["ParamValue"].value == old_value

    results = sim.results
    df = results.to_pandas()
    assert len(df["TestOutput.out_0"]) > 0

    # Check that the constant block has the right value
    expected_result = np.array([new_value] * len(df["TestOutput.out_0"]))
    assert np.isclose(df["TestOutput.out_0"], expected_result).all()

    # Set param with method
    new_value = random.random()
    model.set_parameters({"ParamValue": new_value})

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"

    # Check the parameter value did not change: parameters aren't supposed to be changed
    model.reload()
    assert model.parameters["ParamValue"].value == old_value

    results = sim.results
    df = results.to_pandas()
    assert len(df["TestOutput.out_0"]) > 0

    # Check that the constant block has the right value
    expected_result = np.array([new_value] * len(df["TestOutput.out_0"]))
    assert np.isclose(df["TestOutput.out_0"], expected_result).all()

    # Set param and save to model
    new_value = random.random()
    model.set_parameters({"ParamValue": new_value}, save=True)

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"

    # Check the parameter value updated
    model.reload()
    assert model.parameters["ParamValue"].value == new_value

    results = sim.results
    df = results.to_pandas()
    assert len(df["TestOutput.out_0"]) > 0

    # Check that the constant block has the right value
    expected_result = np.array([new_value] * len(df["TestOutput.out_0"]))
    assert np.isclose(df["TestOutput.out_0"], expected_result).all()


@pytest.mark.parametrize("name", ["test_005"])
def test_configs(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    # Set a config with property setter
    original_stop_time = model.configuration.stop_time
    new_stop_time_1 = original_stop_time + 1
    model.configuration.stop_time = new_stop_time_1
    assert model.configuration.stop_time == new_stop_time_1

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"
    assert sim.results.to_pandas().index[-1] == new_stop_time_1

    # Set a config with method
    new_stop_time_2 = original_stop_time + 2
    model.set_configuration({"stop_time": new_stop_time_2})
    assert model.configuration.stop_time == new_stop_time_2

    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "completed"
    assert sim.results.to_pandas().index[-1] == new_stop_time_2

    # Model configs are preserved by default
    original_model = collimator.load_model(model_name)
    assert original_model.configuration.stop_time == original_stop_time


@pytest.mark.parametrize("name", ["test_002"])
def test_stop_simulation(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    sim = collimator.run_simulation(model, wait=False)
    sim.stop()


if __name__ == "__main__":
    sys.exit(pytest.main([__file__]))



from typing import List
import numpy as np
import ipywidgets as widgets
from IPython.display import display
from pycollimator.models import load_model
from pycollimator.simulations import Simulation, run_simulation


class _Widget:
    """Our representation of ipywidgets to manage shared state, and control results and displays.
    Seems hacky but was the fastest way. Should not be directly instantiated."""

    def __init__(self):
        self._model = None
        self._results = None
        self._widget = None
        # TODO add results plotting widget and store

    @property
    def results(self) -> List[Simulation]:
        """The simulations run by the widget

        :return: List of Simulations run by the widget
        :rtype: List[Simulation]
        """
        return self._results

    def display(self):
        """Display the widget"""
        display(self._widget)


def parameter_sweep_widget():
    """Create a parameter sweep widget.

    :return: A Widget instance that can be displayed and read from with the `results` field.
    :rtype: _Widget
    """
    widget = _Widget()

    param = widgets.Dropdown(description="Parameter")
    start = widgets.FloatText(description="Start")
    stop = widgets.FloatText(description="Stop", value=10)
    step = widgets.FloatText(description="Step", value=0.5)

    def run_parameter_sweep(param, start, stop, step):
        sim_data = list()
        for value in np.arange(start, stop, step):
            sim_data.append(run_simulation(widget._model, {param: value}, wait=False))
        widget._results = sim_data

    parameter_sweep = widgets.interactive(
        run_parameter_sweep,
        {"manual": True, "manual_name": "Run Parameter Sweep"},
        param=param,
        start=start,
        stop=stop,
        step=step,
    )

    def load_model_params(model):
        widget._model = load_model(model)
        param.options = widget._model.parameters.keys()

    model_select = widgets.interactive(
        load_model_params,
        {"manual": True, "manual_name": "Load Model"},
        model=widgets.Text(placeholder="Enter model name", description="Model"),
    )

    widget._widget = widgets.VBox([model_select, parameter_sweep])

    return widget



import sys

import numpy as np
import pandas as pd
import pytest


import pycollimator as collimator
from pycollimator.diagrams import DataSourceBlock

from tests import get_model_name


@pytest.mark.parametrize("name", ["test_003"])
def test_datasource_simple(name: str):
    model_name = get_model_name(name)
    model = collimator.load_model(model_name)

    # Check that the simulation can't run without a valid datasource input
    sim = collimator.run_simulation(model, wait=True)
    assert sim.status.lower() == "failed"

    # 10+1 random points at 0.1s intervals
    test_data = np.random.rand(11)
    test_times = np.linspace(0, 1.0, len(test_data))
    test_df = pd.DataFrame(test_data, index=test_times, columns=["value"])

    block: DataSourceBlock = model.find_block("DataSourceBlock")
    assert isinstance(block, DataSourceBlock)

    block_sampling_interval = block.get_parameter("sampling_interval")
    print("block sampling interval", block_sampling_interval, type(block_sampling_interval))
    assert np.isclose(block_sampling_interval, 0.1)

    print("block data set to", test_df)
    block.set_data(test_df)

    # Interpolation is set to a higher value that
    configuration = {"stop_time": 1.0, "discrete_step": 0.1, "interpolation": 10.0}
    model.set_configuration(configuration)

    sim = collimator.run_simulation(model, wait=True)
    df = sim.results.to_pandas()

    # TestOutput is a ZOH block used to ensure that the DataSource block is read as a discrete block
    print(df)
    assert len(df["TestOutput.out_1"]) > 0
    assert np.isclose(df["TestOutput.out_1"], test_data).all()


if __name__ == "__main__":
    sys.exit(pytest.main([__file__]))



import mimetypes
import os
import requests
import typing as T
from uuid import uuid4

from pycollimator.i18n import N
from pycollimator.log import Log
from pycollimator.error import CollimatorAuthenticationError, CollimatorApiError, CollimatorRetryableApiError
from pycollimator.global_variables import GlobalVariables


class Api:
    BASE_PATH = "/api/v0"

    @classmethod
    def _handle_error(cls, response: T.Any, headers: T.Dict[str, str] = None):
        request_id = (headers is not None) and headers.get("X-Request-ID") or None
        # correlation_id = headers and headers.get('X-Correlation-ID') or None
        if response.status_code >= 400:
            if response.status_code == 401:
                Log.debug("API error 401, response:", response.text)
                raise CollimatorAuthenticationError(
                    N("Authentication failed. Please check your token or reload the page.")
                )
            elif response.status_code == 504:
                Log.trace("API error 504, retrying...")
                raise CollimatorRetryableApiError
            else:
                Log.debug("API error, status_code:", response.status_code, "response:", response.text)
                raise CollimatorApiError(
                    f"API call failed (status: {response.status_code}). Support request ID: "
                    + f"{request_id}. Detailed error:\n{response.text}"
                )

    # TODO add streaming support for logs and results
    @classmethod
    def _call_api(
        cls, api, method, body: T.Any = None, headers: dict = None, body_type="json", response_type="json", retries=3
    ):
        try:
            url = GlobalVariables.url() + api
            headers = headers or GlobalVariables.custom_headers()
            headers["X-Collimator-API-Caller"] = "pycollimator"
            headers["X-Request-ID"] = str(uuid4())
            if body is None:
                Log.trace(method, url)
                response = requests.request(method, url, headers=headers)
            elif body_type == "json":
                Log.trace(method, url, body)
                headers["Content-Type"] = "application/json"
                response = requests.request(method, url, json=body, headers=headers)
            elif body_type == "files":
                Log.trace(method, url, body)
                response = requests.request(method, url, files=body, headers=headers)
            cls._handle_error(response, headers)
            Log.trace(method, "response:", response.text)
            Log.trace(method, "response:", response.status_code)
            if response.status_code != requests.codes["no_content"]:
                if response_type == "text":
                    return response.text
                elif response.headers.get("content-type") and response.headers["content-type"].strip().startswith(
                    "application/json"
                ):
                    return response.json()
                else:
                    raise CollimatorApiError(
                        f"Unexpected response type (status: {response.status_code}). Support request ID: "
                        + f"{headers['X-Request-ID'] }. Detailed error:\n{response.text}"
                    )

        except CollimatorRetryableApiError as e:
            if retries > 0:
                Log.trace(method, "retrying...")
                return cls._call_api(api, method, body, headers, body_type, response_type, retries - 1)
            raise e

    @classmethod
    def get(cls, api):
        return cls._call_api(api, method="GET")

    @classmethod
    def post(cls, api, body):
        return cls._call_api(api, method="POST", body=body)

    @classmethod
    def put(cls, api, body):
        return cls._call_api(api, method="PUT", body=body)

    @classmethod
    def delete(cls, api, body):
        return cls._call_api(api, method="DELETE", body=body)

    @classmethod
    def get_projects(cls):
        api = f"{Api.BASE_PATH}/projects"
        return cls.get(api)

    @classmethod
    def get_project(cls, project_uuid: str = None) -> dict:
        project_uuid = project_uuid or GlobalVariables.project_uuid()
        api = f"{Api.BASE_PATH}/projects/{project_uuid}"
        return cls.get(api)

    @classmethod
    def get_project_by_name(cls, name: str) -> dict:
        projects = cls.get_projects()["projects"]
        for p in projects:
            if p["title"] == name:
                return p
        return {}

    @classmethod
    def get_model(cls, model_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}"
        return cls.get(api)

    @classmethod
    def simulation_create(cls, model_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations"
        return cls.post(api, body)

    @classmethod
    def simulation_get(cls, model_uuid: str, simulation_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}"
        return cls.get(api)

    @classmethod
    def simulation_start(cls, model_uuid: str, simulation_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/start"
        return cls.post(api, None)

    @classmethod
    def simulation_stop(cls, model_uuid: str, simulation_uuid: str) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/events"
        return cls.post(api, {"command": "stop"})

    @classmethod
    def simulation_parameters_set(cls, model_uuid: str, simulation_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/parameters"
        return cls.put(api, body)

    @classmethod
    def simulation_update(cls, model_uuid: str, simulation_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}"
        return cls.put(api, body)

    @classmethod
    def simulation_hashed_file_create(cls, model_uuid: str, simulation_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/hashed_files"
        return cls.post(api, body)

    @classmethod
    def simulation_hashed_file_upload(cls, path: str, upload_url: str) -> dict:
        with open(path, "rb") as file:
            return cls._call_api(upload_url, method="PUT", body={"file": file}, body_type="files")

    # FIXME should be a stream
    @classmethod
    def simulation_logs(cls, model_uuid: str, simulation_uuid: str) -> str:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/logs"
        return cls._call_api(api, method="GET", response_type="text")

    # FIXME should be a stream
    @classmethod
    def simulation_results(cls, model_uuid: str, simulation_uuid: str, retries=3) -> str:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}"
        api += "/process_results?files=continuous_results.csv"

        s3_urls_batch_resp = cls._call_api(api, method="GET", retries=retries)
        urls = cls._get_result_urls(s3_urls_batch_resp)

        return cls._get_file(urls[0])

    @classmethod
    def signal_results(cls, model_uuid: str, simulation_uuid: str, paths: T.List[str], retries=3) -> T.List[str]:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}"
        api += f"/process_results?files={','.join(paths)}"

        s3_urls_batch_resp = cls._call_api(api, method="GET", retries=retries)
        urls = cls._get_result_urls(s3_urls_batch_resp)

        # TODO: async fetch if users start specifying many signals
        return [cls._get_file(url) for url in urls]

    # @classmethod
    # def model_configuration_update(cls, model_uuid: str, body: dict) -> dict:
    #     api = f"{Api.BASE_PATH}/models/{model_uuid}/configuration"
    #     return cls.put(api, body)

    @classmethod
    def model_parameter_update(cls, model_uuid: str, body: dict) -> dict:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/parameters"
        return cls.put(api, body)

    @classmethod
    def linearization_results_csv(cls, model_uuid: str, simulation_uuid: str) -> str:
        api = f"{Api.BASE_PATH}/models/{model_uuid}/simulations/{simulation_uuid}/lin_results"
        return cls._call_api(api, method="GET", response_type="text")

    @classmethod
    def upload_file(cls, filepath: str, project_uuid: str = None, overwrite: bool = False) -> str:
        project_uuid = project_uuid or GlobalVariables.project_uuid()
        content_type, _ = mimetypes.guess_type(filepath)
        size = os.stat(filepath).st_size

        response = cls._call_api(
            f"{Api.BASE_PATH}/projects/{project_uuid}/files",
            method="POST",
            body={
                "name": os.path.basename(filepath),
                "content_type": content_type or "application/octet-stream",
                "size": size,
                "overwrite": overwrite,
            },
        )

        file_uuid = response["summary"]["uuid"]
        s3_url = response["put_presigned_url"]

        with open(filepath, "rb") as f:
            response = requests.put(s3_url, data=f, headers={"Content-Type": content_type, "Content-Length": str(size)})

        if response.status_code != 200:
            Log.debug("API error, status_code:", response.status_code, "response:", response.json)
            raise CollimatorApiError(f"Failed to upload to S3. Detailed error:\n{response.text}")

        return cls._call_api(f"{Api.BASE_PATH}/projects/{project_uuid}/files/{file_uuid}/process", method="POST")

    @staticmethod
    def _get_result_urls(batch_resp: str) -> T.List[str]:
        errors = []
        urls = []

        for s3_url_resp in batch_resp["s3_urls"]:
            url = s3_url_resp.get("url")
            if url is None:
                errors.append(f"Error for signal '{s3_url_resp.get('name')}'\n{s3_url_resp.get('error')}")
            else:
                urls.append(url)

        if len(errors) > 0:
            raise CollimatorApiError("\n".join(errors))

        return urls

    @classmethod
    def _get_file(cls, url: str) -> str:
        response = requests.request("GET", url)
        cls._handle_error(response)
        return response.text



import traceback

# import sys

from pycollimator.i18n import N


def quiet_hook(kind, message, tb):
    traceback.print_tb(tb, limit=1)
    print("{0}: {1}".format(kind.__name__, message))  # Only print Error Type and Message


# sys.excepthook = quiet_hook


class Error(Exception):
    pass


class NoModelError(Error):
    """No model matches the string given"""

    def __init__(self, name):
        message = "There are no models with the name " + name + ".  Please check your model editor and try again."
        super().__init__(message)


class NoModelsError(Error):
    """No models are part of this project"""

    def __init__(self):
        message = "There are no models in this project.  Please add one to load in the notebook."
        super().__init__(message)


class NoModelParamsError(Error):
    def __init__(self):
        message = "There are no model parameters for this model.  Please add one through the model editor."
        super().__init__(message)


class ModelNotValidError(Error):
    def __init__(self):
        message = "This model is not valid for simulation.  Please correct the model through the model editor."
        super().__init__(message)


class NotFoundError(Error):
    def __init__(self, message: str = None):
        super().__init__(message or N("Not found"))


class NotLoadedError(Error):
    def __init__(self, message: str = None):
        super().__init__(message or N("Not loaded"))


class UnsupportedOperationError(Error):
    def __init__(self, message: str = None):
        super().__init__(message or N("Unsupported operation"))


class CollimatorApiError(Error):
    def __init__(self, message: str = None):
        super().__init__(message or N("Collimator API Error"))


class CollimatorRetryableApiError(Error):
    def __init__(self, message: str = None):
        super().__init__(message or N("Collimator API Error"))


class CollimatorRuntimeError(Error):
    def __init__(self, simulation_status_json):
        message = simulation_status_json["fail_reason"] + "Please return to model editor to correct."
        super().__init__(message)


class CollimatorAuthenticationError(Error):
    def __init__(self, message):
        message = message or "Jupyter authentication has expired.  Please reload jupyter to reauthenticate user."
        super().__init__(message)


class CollimatorUnexpectedError(Error):
    def __init__(self, msg: str = None):
        message = N("An unexpected error occured") + (msg is None and "." or ": ") + msg
        super().__init__(message)



import re
import typing as T

from pycollimator.log import Log


class GlobalVariables:
    """
    global variables stores information about user's authentication and the project
    folder they are currently in
    """

    _instance: "GlobalVariables" = None
    _envRegex = re.compile("https://(dev|test|app).collimator.ai")

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(GlobalVariables, cls).__new__(cls)
        return cls._instance

    def set_auth_token(self, token: str, project: str = None, api_url: str = None) -> None:
        if api_url is None:
            api_url = "https://app.collimator.ai"
        self.auth_token = token
        self.project = project
        self.api_url = api_url
        Log.trace("auth_token:", self.auth_token, "project:", self.project, "api_url:", self.api_url)

    @classmethod
    def _get_instance(cls) -> "GlobalVariables":
        if cls._instance is None:
            cls()
        return cls._instance

    @classmethod
    def project_uuid(cls):
        """
        stores the project uuid associated with the folder.
        """
        return cls._get_instance().project

    @classmethod
    def token(cls):
        """
        stores the authentication token
        """
        return cls._get_instance().auth_token

    @classmethod
    def url(cls):
        """
        stores the url, used for environment logic
        """
        return cls._get_instance().api_url

    @classmethod
    def custom_headers(cls) -> T.Dict[str, str]:
        """
        stores the authentication headers used in all API requests
        """
        # FIXME I think requests should be handling the content-type header for us
        custom_headers = {
            "X-Collimator-API-Token": cls._get_instance().auth_token,
            "Accept": "application/json",
            # "content-type": "application/json",
        }
        return custom_headers


def get_project_url() -> str:
    return GlobalVariables.url() + "/projects/" + GlobalVariables.project_uuid()


def set_auth_token(token: str, project_uuid: str = None, api_url: str = None) -> None:
    GlobalVariables._get_instance().set_auth_token(token, project_uuid, api_url)



from os import environ


class Log:
    log_level = environ.get("LOG_LEVEL", "INFO")

    @classmethod
    def __should_log(cls, level: str):
        app_level = Log.log_level or "INFO"
        level_map = {
            "TRACE": 5,
            "DEBUG": 4,
            "INFO": 3,
            "WARNING": 2,
            "ERROR": 1,
            "FATAL": 0,
        }
        return level_map.get(app_level, 2) >= level_map.get(level, 4)

    @classmethod
    def trace(cls, *args, **kwargs):
        if not Log.__should_log("TRACE"):
            return
        print("TRACE:", *args, **kwargs)

    @classmethod
    def debug(cls, *args, **kwargs):
        if not Log.__should_log("DEBUG"):
            return
        print("DEBUG:", *args, **kwargs)

    @classmethod
    def info(cls, *args, **kwargs):
        if not Log.__should_log("INFO"):
            return
        print("INFO:", *args, **kwargs)

    @classmethod
    def warning(cls, *args, **kwargs):
        if not Log.__should_log("WARNING"):
            return
        print("WARNING:", *args, **kwargs)

    @classmethod
    def error(cls, *args, **kwargs):
        if not Log.__should_log("ERROR"):
            return
        print("ERROR:", *args, **kwargs)

    @classmethod
    def fatal(cls, *args, **kwargs):
        if not Log.__should_log("FATAL"):
            return
        print("FATAL:", *args, **kwargs)

    @classmethod
    def set_level(cls, level: str):
        if level in ["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "FATAL"]:
            Log.log_level = level
        else:
            raise ValueError(f"Invalid log level: {level}")

    @classmethod
    def is_level_above(cls, lvl="INFO"):
        return cls.__should_log(lvl)
